import warnings


def warning_on_one_line(message, category, filename, lineno, file=None, line=None):
    return ' %s:%s: %s: %s\n' % (filename, lineno, category.__name__, message)


warnings.formatwarning = warning_on_one_line


class MissingAtomWarning(Warning):
    pass


class DifferentVerteciesWarning(Warning):
    pass


class EmptyLineWarning(Warning):
    pass
